
-- # import QuickCheck
-- # import Bool.Proofs
-- # import Ord
-- # import Maybe.Proofs
-- # import List.Proofs
-- # import Nat.Proofs

data Heap (A : Set) : Set where
  Node	: A -> Heap A -> Heap A -> Heap A
  Empty	: Heap A

empty : (A : Set) -> Heap A
empty A = Empty

isEmpty : (A : Set) -> Heap A -> Bool
isEmpty A Empty		= true
isEmpty A (Node x l r)  = false

unitHeap : (A : Set) -> A -> Heap A
unitHeap A x = Node x Empty Empty

size : (A : Set) -> Heap A -> Nat
size A Empty = 0
size A (Node x l r) = size A l + size A r + 1

merge : (A : Set) -> Ord A -> Heap A -> Heap A -> Heap A
merge A ordA h1 Empty = h1
merge A ordA Empty h2 = h2
merge A ordA (Node x l1 r1) (Node y l2 r2) =
    if (ordA.leq x y)
       (Node x (merge A ordA r1 (Node y l2 r2)) l1)
       (Node y (merge A ordA r2 (Node x l1 r1)) l2)

insertHeap : (A : Set) -> Ord A -> A -> Heap A -> Heap A
insertHeap A ordA x h = merge A ordA (unitHeap A x) h

removeMin : (A : Set) -> Ord A -> Heap A -> Maybe {A ; Heap A}
removeMin A ordA Empty	      = nothing
removeMin A ordA (Node x l r) = just (x, merge A ordA l r)

-- Helpers to fromList (not structurally recursive!)
sweep : (A : Set) -> Ord A -> List (Heap A) -> List (Heap A)
sweep A ordA []		= []
sweep A ordA [h]	= [h]
sweep A ordA (h1::h2::hs) = merge A ordA h1 h2 :: sweep A ordA hs

merging : (A : Set) -> Ord A -> List (Heap A) -> Heap A
merging A ordA []  = empty A
merging A ordA [h] = h
merging A ordA hs  = merging A ordA (sweep A ordA hs)

fromList : (A : Set) -> Ord A -> List A -> Heap A
fromList A ordA xs = merging A ordA (map (unitHeap A) xs)

-- Not structurally recursive!
-- Would the termination checker handle this?
toList' : (A : Set) -> List (Heap A) -> List A
toList' A []		      = []
toList' A (Empty :: hs)	      = toList' A hs
toList' A (Node x h1 h2 :: hs) = x :: toList' A (h1::h2::hs)

toList : (A : Set) -> Heap A -> List A
toList A h = toList' A [h]

toSortedList : (A : Set) -> Ord A -> Heap A -> List A
toSortedList A ordA Empty = []
toSortedList A ordA (Node x h1 h2) = x :: toList A (merge A ordA h1 h2)
-- The above line contains a bug. The correct version is::
-- toSortedList A ordA (Node x h1 h2) = x :: toSortedList A ordA (merge A ordA h1 h2)

-- Generators -------------------------------------------------------------

tries : (A : Set) -> Gen A -> (A -> Bool) -> Nat -> Gen (Maybe A)
tries A gen p 0 = monadGen.return (Maybe A) nothing
tries A gen p (suc n) =
    monadGen.bind A (Maybe A) gen
	(\a -> if (p a)
		  (monadGen.return (Maybe A) (just a))
		  (tries A gen p n)
	)

suchThatMaybe : (A : Set) -> Gen A -> (A -> Bool) -> Gen (Maybe A)
suchThatMaybe A g p = sized' (Maybe A) (tries A g p)

arbHeap' : (A : Set) -> Ord A -> Gen A -> Maybe A -> Nat -> Gen (Heap A)
arbHeap' A ordA arbA mx 0 = monadGen.return (Heap A) Empty
arbHeap' A ordA arbA mx n =
    frequency (Heap A)
	[ (1, monadGen.return (Heap A) Empty)
	, (7, monadGen.bind (Maybe A) (Heap A)
		(suchThatMaybe A arbA (\y -> (ordMaybe A ordA).leq mx (just y)))
		(maybe
		  (monadGen.return (Heap A) Empty)
		  (\y -> let arbHeap2 = arbHeap' A ordA arbA (just y) (half n)
			 in liftM2 Gen monadGen (Heap A) (Heap A) (Heap A)
				   (Node y) arbHeap2 arbHeap2
		  )
		)
	  )
	]

arbHeap : (A : Set) -> Ord A -> Gen A -> Gen (Heap A)
arbHeap A ordA arbA = sized' (Heap A) (\n -> arbHeap' A ordA arbA nothing (half n))

-- Properties -------------------------------------------------------------

leqHeap : (A : Set) -> Ord A -> A -> Heap A -> Bool
leqHeap A ordA x Empty	      = true
leqHeap A ordA x (Node y l r) = ordA.leq x y

invariant : (A : Set) -> Ord A -> Heap A -> Bool
invariant A ordA Empty = true
invariant A ordA (Node x l r) =
    x <=? l && x <=? r && invariant A ordA l && invariant A ordA r
  where
    (<=?) = leqHeap A ordA

-- We don't want to use boolean list equality (it's slow)...
-- ...but we have to, since all comparisons are using the
-- boolean equality (using Eq results in a too strong property).
eqHeap : (A : Set) -> Ord A -> Heap A -> List A -> Bool
eqHeap A ordA h xs = invariant A ordA h &&
		      (equList A ordA.equ).eq
			  (sortList ordA (toList A h))
			  (sortList ordA xs)

-- empty
PropEmpty : (A : Set) -> Ord A -> Prop
PropEmpty A ordA = Holds (eqHeap A ordA (empty A) [])

propEmpty : (A : Set) -> (ordA : Ord A) -> PropEmpty A ordA
propEmpty A ordA = tt

--testEmpty : PropEmpty Nat ordNat
--testEmpty = qc-plugin()

-- unitHeap
PropUnit : (A : Set) -> Ord A -> Prop
PropUnit A ordA = (x : A) -> Holds (eqHeap A ordA (unitHeap A x) [x])

propUnit : (A : Set) -> (ordA : Ord A) -> PropUnit A ordA
propUnit A ordA = fol-plugin(propHoldsAndI, ordA.equ.proofs.refl, Holds)

--testUnit : PropUnit Nat ordNat
--testUnit = qc-plugin(arbNat)

-- size
PropSize : (A : Set) -> Prop
PropSize A = (h : Heap A) -> size A h == length (toList A h)

propLengthToList : (A : Set) -> (hs : List (Heap A)) ->
		   (sum (map (size A) hs)) ==
			   (length (toList' A hs))
propLengthToList A []			= reflexivity 0
propLengthToList A (Empty :: hs)		=
    fol-plugin(propPlusZero, propLengthToList A hs)
propLengthToList A (Node x h1 h2 :: hs)	=
    fol-plugin(	propPlusSuc
	      , propLengthToList A (h1::h2::hs)
	      , propPlusAssoc
	      )

propSize : (A : Set) -> PropSize A
propSize A h = propLengthToList A [h]

--testSize : PropSize Bool
--testSize = qc-plugin(arbHeap Bool ordBool arbBool)

-- merge
PropMerge : (A : Set) -> Ord A -> Prop
PropMerge A ordA =
  (h1, h2 : Heap A) ->
    Holds (invariant A ordA h1) ==>
    Holds (invariant A ordA h2) ==>
    Holds (
      eqHeap A ordA (merge A ordA h1 h2)
		    (toList A h1 ++ toList A h2)
    )

propMerge : (A : Set) -> (ordA : Ord A) -> PropMerge A ordA
propMerge A ordA h1 Empty =
    fol-plugin(	(equList A ordA.equ).proofs.refl
	      , propAppNil A
	      , propHoldsAndI
	      )
propMerge A ordA Empty (Node x l r) =
    fol-plugin( (equList A ordA.equ).proofs.refl
	      , propHoldsAndI
	      )
propMerge A ordA (Node x l1 r1) (Node y l2 r2) = assert-plugin()
{- ** should be converted (using impE or impI) between ==> and ->
    \inv1 inv2 ->
      let h1 : Heap A; h1 = Node x l1 r1
	  h2 : Heap A; h2 = Node y l2 r2
	  rec1 = propMerge A ordA r1 h2
	  rec2 = propMerge A ordA r2 h1
      in
      propIf (Heap A) (\z -> Holds (invariant A ordA z &&
				    eqList A ordA.equ
				      (sortList ordA (toList A z))
				      (sortList ordA 
					(toList A h1 ++ toList A h2)
				      )
				    )
		      )
	(ordA.leq x y)
	(Node x (merge A ordA r1 h2) l1)
	(Node y (merge A ordA r2 h1) l2)
	(\x_leq_y ->

	  let rem1  : Holds (leqHeap A ordA x (merge A ordA r1 h2))
	      rem1 = assert-plugin(silent)

	      rem2  : Holds (leqHeap A ordA x l1)
	      rem2 = fol-plugin(inv1, propHoldsAndE)

	      rem3  : Holds (invariant A ordA (merge A ordA r1 h2))
	      rem3 = fol-plugin(rec1, inv1, inv2, propHoldsAnd)

	      rem4  : Holds (invariant A ordA l1)
	      rem4 = fol-plugin(inv1, propHoldsAndE)

	      rem5  : Holds
			(eqList A ordA.equ
			  (sortList ordA
			    (toList A (Node x (merge A ordA r1 (Node y l2 r2)) l1))
			  )
			  (sortList ordA
			    (toList A (Node x l1 r1) ++ 
				      toList A (Node y l2 r2)
			    )
			  )
			)
	      rem5 = ?

	      -- Takes ages for the fol-plugin. Why?
	      rem6 : (a,b,c,d,e : Bool) ->
		      Holds a -> Holds b -> Holds c -> Holds d -> Holds e ->
		      Holds ((a && b && c && d) && e)
	      rem6 a b c d e pa pb pc pd pe =
		  propHoldsAndI (a && b && c && d) e
		    ( propHoldsAndI a (b && c && d)
		      ( pa
		      , propHoldsAndI b (c && d)
			  ( pb
			  , propHoldsAndI c d (pc, pd)
			  )
		      )
		    , pe
		    )

	      debug1 = debug-plugin(inv1)
	      debug2 = debug-plugin(inv2)
	      debug3 = debug-plugin(x_leq_y)
	      debug4 = debug-plugin(propMerge A ordA r1 h2)
	      --debug5 = debug-plugin(propMerge A ordA r2 h1)

	  in  fol-plugin( rem1, rem2, rem3
			, rem4, rem5, rem6
			)
	)
	(\x_gt_y -> assert-plugin(silent))
-}

--testMerge : PropMerge Nat ordNat
--testMerge = qc-plugin(arbHeap Nat ordNat arbNat, arbHeap Nat ordNat arbNat)

-- insert
PropInsert : (A : Set) -> Ord A -> Prop
PropInsert A ordA = (x : A) -> (h : Heap A) ->
		      Holds (invariant A ordA h) ==>
		      Holds (
			eqHeap A ordA
			  (insertHeap A ordA x h)
			  (x :: toList A h)
		      )

propInsert : (A : Set) -> (ordA : Ord A) -> PropInsert A ordA
propInsert A ordA x h = impI (\invh -> 
                              propMerge A ordA (unitHeap A x) h 
                                 `impE` tt `impE` invh)

-- *very* slow, maybe I should do some work on the compiler...
--testInsert : PropInsert Bool ordBool
--testInsert = qc-plugin(arbBool, arbHeap Bool ordBool arbBool)

-- removeMin
prop_removeMin' : (A : Set) -> Ord A -> Heap A -> Maybe {A ; Heap A} -> Bool
prop_removeMin' A ordA h nothing       = eqHeap A ordA h []
prop_removeMin' A ordA h (just (x,h')) =
      ordA.equ.eq x (minimum ordA (toList A h)) &&
      eqHeap A ordA h' (delete ordA.equ x (toList A h))

PropRemoveMin : (A : Set) -> Ord A -> Prop
PropRemoveMin A ordA =
  (h : Heap A) ->
    Holds (invariant A ordA h) ==>
    Holds (prop_removeMin' A ordA h (removeMin A ordA h))

--testRemoveMin : PropRemoveMin Nat ordNat
--testRemoveMin = qc-plugin(arbHeap Nat ordNat arbNat)

-- fromList
PropFromList : (A : Set) -> Ord A -> Prop
PropFromList A ordA = (xs : List A) ->
		      Holds (eqHeap A ordA (fromList A ordA xs) xs)

--testFromList : PropFromList Bool ordBool
--testFromList = qc-plugin(arbList Bool arbBool)

-- toSortedList
PropToSortedList : (A : Set) -> Ord A -> Prop
PropToSortedList A ordA =
    (h : Heap A) ->
    Holds (invariant A ordA h) ==>
    let xs = toSortedList A ordA h
    in Holds (eqHeap A ordA h xs && ((equList A ordA.equ).eq xs (sortList ordA xs)))

-- Finds the bug quite quickly
--testToSortedList : PropToSortedList Nat ordNat
--testToSortedList = qc-plugin(arbHeap Nat ordNat arbNat)

