
\section{The QuickCheck Library} \label{QuickCheck.lagda}

\begin{code}
-- # import Property
-- # import Monad
-- # import Monad.Reader
-- # import Nat
-- # import List
\end{code}

The QuickCheck library defines functions to build the generators used to
generate the test cases, and provides special property constants to 

%- Random numbers ---------------------------------------------------------
\subsection{Producing Random Values}

\begin{code}
StdGen : Set

Rand : (A : Set) -> Set
Rand A = StdGen -> A

split : Rand { StdGen ; StdGen }

monadR : MonadOps Rand
monadR = struct return = return; bind = bind
  where
    return : (A : Set) -> A -> Rand A
    return = \A x r -> x

    bind : (A, B : Set) -> Rand A -> (A -> Rand B) -> Rand B
    bind = \A B m k r -> let rs = split r
                         in k (m rs.1) rs.2

seed  : StdGen
-- What if the list is empty? No problem, just return epsilon A!
pick  : (A : Set) -> List A -> Rand A

CanProduce : (A:Set) -> Rand A -> A -> Prop
CanProduce A f x = { r : StdGen; f r == x }

step : Nat -> StdGen -> StdGen
step 0 r       = r
step (suc n) r = step n (split r).1
\end{code}

%- Generators -------------------------------------------------------------
\subsection{Building Test Data Generators}

\begin{code}
Gen : (A : Set) -> Set
Gen A = Nat -> StdGen -> A

CanGenerate : (A : Set) -> Gen A -> A -> Prop
CanGenerate A g x = { size : Nat; CanProduce A (g size) x }

-- primitives
-- Gen is not a proper monad.
monadGen : MonadOps Gen
monadGen = (monadTransReaderT Nat).ops.monad Rand monadR

sized : (A : Set) -> (Nat -> Gen A) -> Gen A
sized A fgen n r = fgen n n r

rand : Gen StdGen
rand n r = r

promote : (A, B : Set) -> (A -> Gen B) -> Gen (A -> B)
promote A B f n r = \a -> f a n r

variant : (A : Set) -> Nat -> Gen A -> Gen A
variant A v m n r = m n (step (suc n) r)

-- derived

elements : (A : Set) -> List A -> Gen A
elements A xs = liftM Gen monadGen StdGen A
                  (pick A xs) rand

oneof : (A : Set) -> List (Gen A) -> Gen A
oneof A gens = join Gen monadGen A (elements (Gen A) gens)

Bag : Set -> Set
Bag A = List { Nat ; A }

flattenBag : (A : Set) -> Bag A -> List A
{-
-- Problem - postulate if for now.
flattenBag A ((0,     x) :: xs) = flattenBag A xs
flattenBag A ((suc n, x) :: xs) = x :: flattenBag A ((n, x) :: xs)
flattenBag A []                 = []
-}

frequency : (A : Set) -> Bag (Gen A) -> Gen A
frequency A xs = oneof A (flattenBag (Gen A) xs)
\end{code}

%- Standard generators ----------------------------------------------------
\subsection{Some Standard Generators}

\begin{code}
-- Independent functions
arbFun : (A, B : Set) -> (A -> Gen B -> Gen B) -> Gen B -> Gen (A -> B)
arbFun A B coarb genb = promote A B (\x -> coarb x genb)

-- Sigma types
arbSigma : (A:Set) -> (B:A -> Set) -> Gen A -> ((x:A) -> Gen (B x)) ->
            Gen {x:A; B x}
arbSigma A B gA fB =  monadGen.bind A     {z:A; B z} gA      (\x ->
                      monadGen.bind (B x) {z:A; B z} (fB x)  (\y ->
                        monadGen.return   {z:A; B z} (x,y)
                      ))

-- Natural numbers
arbNatS : Nat -> Gen Nat
arbNatS 0        = monadGen.return Nat 0
arbNatS (suc n)  =
  frequency Nat
            [ (1, monadGen.return Nat 0)
            , (5, liftM Gen monadGen Nat Nat suc (arbNatS n))
            ]

arbNat : Gen Nat
arbNat = sized Nat arbNatS

coarbNat : (A:Set) -> Nat -> Gen A -> Gen A
coarbNat A n g = variant A n g

-- Booleans
arbBool : Gen Bool
arbBool = elements Bool [false, true]

coarbBool : (A:Set) -> Bool -> Gen A -> Gen A
coarbBool A false g = variant A 0 g
coarbBool A true g  = variant A 1 g

-- Lists
arbListS : (A : Set) -> Gen A -> Nat -> Gen (List A)
arbListS A genA 0         = monadGen.return (List A) []
arbListS A genA (suc n)  =
  frequency (List A)
            [ (1, monadGen.return (List A) [])
            , (5, liftM2 Gen monadGen A (List A) (List A)
                    (::) genA (arbListS A genA n))
            ]

arbList : (A : Set) -> Gen A -> Gen (List A)
arbList A g = sized (List A) (arbListS A g)
\end{code}

%- Property constants -----------------------------------------------------
\subsection{Special Property Constants}

\begin{code}
forAll : (A : Set) -> Gen A -> (A -> Prop) -> Prop
forAll A g f = (x:A) -> CanGenerate A g x -> f x

collect   : (A : Set) -> A -> Prop -> Prop
collect _ _ p = p

label     : (A : Set) -> A -> Prop -> Prop
label _ _ p = p

classify  : (A : Set) -> Bool -> A -> Prop -> Prop
classify _ _ _ p = p

trivial   : Bool -> Prop -> Prop
trivial _ p = p
\end{code}

