{
{-| There are three shift\/reduce conflicts:

    > \x -> x + y

    can be parsed (incorrectly) as @(\\x -> x) + y@ or
    (correctly) as @\\x -> x + y@. The parser will @shift@ and hence do the
    right thing. The same problem occurs for @`op`@ and @->@ in place of @+@.
-}
module Syntax.Parser (
    -- * Exported Functions
    parse
    , parseFile
    , parseLiterate
    , parseLiterateFile
    , programParser
    , termParser
{-  REMOVE_IF_HADDOCK
    -- * Private Functions
    , mkLam
    , mkPi
    , mkInfix
    REMOVE_IF_HADDOCK -}
    ) where

import Syntax.Position
import Syntax.ParseMonad
import Syntax.Concrete
import Syntax.Lexer

import Utils.Monads

}

%name termParser Term
%name programParser Program
%tokentype { Token }
%monad { Parser }
%lexer { lexer } { TkEOF }
-- %expect 3

%token
    '('	    { TkLParen $$ }
    ')'	    { TkRParen $$ }
    '\\'    { TkLambda $$ }
    '->'    { TkArrow $$ }
    '_'     { TkUnderscore $$ }
    id	    { TkId (_,$$) }
    int	    { TkInt (_,$$) }
    '::'    { TkDblColon $$ }
    univ    { TkUniverse (_,$$) }
    let	    { TkLet $$ }
    in	    { TkIn $$ }
    where   { TkWhere $$ }
    Set	    { TkSet $$ }
    Prop    { TkProp $$ }
    Type    { TkType $$ }
    infix   { TkInfix $$ }
    infixl  { TkInfixL $$ }
    infixr  { TkInfixR $$ }
    plugin  { TkPlugin (_,$$) }
    '{'	    { TkOpenBrace $$ }
    '}'	    { TkCloseBrace $$ }
    vopen   { TkVOpenBrace $$ }
    vclose  { TkVCloseBrace $$ }
    '='	    { TkEqual $$ }
    ';'	    { TkSemi $$ }
    vsemi   { TkVSemi $$ }
    ','	    { TkComma $$ }
    '`'	    { TkBackQuote $$ }
    '?'	    { TkOp ($$,"?") }
    op	    { TkOp (_,$$) }
    TeX	    { TkTeX $$ }

%%

-- Programs

Program : Init Tops		{ $2 }

-- Definitions

Top :	topen MaybeDefs close	{ $2 }

Tops
    : Top		{ $1 }
    | Top TeX Tops	{ $1 ++ $3 }
    | TeX Tops		{ $2 }

topen :				{% pushCurrentContext }

Definitions
    : '{' MaybeDefs '}'		{ $2 }
    | vopen MaybeDefs close	{ $2 }

close	: vclose	{ () }
	| error		{% popContext }

semi	: ';'		{ $1 }
	| vsemi		{ $1 }

-- Definitions must be separated by semicolons; however, we are quite liberal.
-- For instance, they can be separated by multiple semicolons, and trailing
-- semicolons after the last definition are also allowed.

-- Not anymore, we aren't. (Not necessary with layout?)

MaybeDefs
    :			    { [] }
    | Def		    { [$1] }
    | Def semi MaybeDefs    { $1:$3 }

Def : Var SpaceNames '=' Term Where
			{ mkFunClause $1 $2 $4 $5 }
    | NameOrUnderscore InfixOp NameOrUnderscore '=' Term Where
			{ mkFunClause $2 [$1,$3] $5 $6
			}
    | Var '::' Term			    { TypeSig $1 $3 }
    | infix int CommaOps		    { mkInfix NonAssoc $2 $3 }
    | infixl int CommaOps		    { mkInfix LeftAssoc $2 $3 }
    | infixr int CommaOps		    { mkInfix RightAssoc $2 $3 }

Where
    :					    { [] }
    | where LocalDefs			    { $2 }

LocalDefs : Definitions	{ $1 }

-- Terms

Term
    : '(' CommaNames '::' Term ')' '->' Term
				    { mkPi $2 $4 $7 }
    | Term1 '->' Term		    { Pi "_" $1 $3 }
    | Term1			    { $1 }

Term1
    : Term1 InfixOp Term2	    { InfixApp $1 $2 $3 }
    | Term2			    { $1 }

Term2
    : '\\' SpaceNames1 '->' Term    { mkLam $2 $4 }
    | let LocalDefs in Term	    { Let $2 $4 }
    | Term3			    { $1 }

Term3
    : Term3 Term4		    { App $1 $2 }
    | Term4			    { $1 }

-- Atomic Terms

Term4
    : Var				    { Name $1 }
    | Universe				    { $1 }
    | plugin '(' MaybeTermList ')'          { CallPlugin $1 $3 }
    | '?'				    { CallPlugin "assert" [] }
    | '(' Term ')'			    { Paren $2 }

Universe
    : univ				    { Star (Univ $1) }
    | Prop				    { Prop }
    | Set				    { Star (Univ 0) }
    | Type				    { Star (Univ 1) }

InfixOp
    : '`' id '`'			    { $2 }
    | op				    { $1 }

MaybeTermList
    :                                       { [] }
    | TermList                              { $1 }

TermList
    : Term                                  { [$1] }
    | Term ',' TermList                     { $1 : $3 }

-- Names
-- "_" is a special identifier which may be the name of an abstraction
-- (in Lam and Pi), but it can never be referred to as variable.

SpaceNames
    :		    { [] }
    | SpaceNames1   { $1 }

SpaceNames1
   : NameOrUnderscore		    { [$1] }
   | NameOrUnderscore SpaceNames1   { $1 : $2 }

CommaNames
    : NameOrUnderscore			{ [$1] }
    | NameOrUnderscore ',' CommaNames	{ $1 : $3 }

CommaOps
    : InfixOp		    { [$1] }
    | InfixOp ',' CommaOps  { $1 : $3 }

NameOrUnderscore 
    : '_'                                   { "_" }     -- without Range
    | Name                                  { $1 }

Name : Var				    { $1 }	-- without Range

Var : id				    { $1 }      -- without Range
    | '(' op ')'			    { $2 }

Init : { }

{

-- Parsing

parse :: Parser a -> String -> ParseResult a
parse p = parse' 0 p

parseFile :: Parser a -> FilePath -> IO (ParseResult a)
parseFile p = parseFile' 0 p

parseLiterate :: Parser a -> String -> ParseResult a
parseLiterate p = parse' tex p

parseLiterateFile :: Parser a -> FilePath -> IO (ParseResult a)
parseLiterateFile p = parseFile' tex p

programParser	:: Parser Program
termParser	:: Parser Term

happyError = fail "Parse error"

oneofId xs y
    | elem y xs	= return y
    | otherwise	= fail $ "Expected one of " ++ unwords xs

-- | Even though this function is not exported, I want it to show up
--   in the documentation.
mkLam :: [Name] -> Term -> Term
mkLam vs t = foldr Lam t vs

mkPi :: [Name] -> Term -> Term -> Term
mkPi xs t u = foldr (\x -> Pi x t) u xs

mkInfix :: Assoc -> Precedence -> [Name] -> Decl
mkInfix ass pr vs = Infix ass pr vs

mkFunClause :: Name -> [Name] -> Term -> [Decl] -> Decl
mkFunClause x xs e [] = FunClause x (mkLam xs e)
mkFunClause x xs e ds	= FunClause x (mkLam xs $ Let ds e)

}
