{-| Defines the various error messages emitted by the type checker. -}
module TypeChecker.Errors where

import Text.PrettyPrint

import Syntax.Position
import Syntax.Internal
import Syntax.Pretty
import Utils.Error

-- Whats and whens --------------------------------------------------------

whenCheckingDefinition :: Name -> Doc
whenCheckingDefinition x =
    fsep $ pwords $ "checking the definition of " ++ x

whenCheckingType :: Term -> Value -> Doc
whenCheckingType e v =
    fsep $ pwords "checking that" ++ [pretty e] ++
	    pwords "has type" ++ [pretty v]

whenCheckingConstructor :: Name -> Doc
whenCheckingConstructor c =
    fsep $ pwords "checking the type of the constructor" ++ [text c]


whenInferringType :: Term -> Doc
whenInferringType e =
    fsep $ pwords "inferring the type of" ++ [pretty e]

whenTranslatingToFOL :: Value -> Doc
whenTranslatingToFOL v =
    fsep $ pwords "translating" ++ [pretty v] ++
	   pwords "to first order logic"

whenLoadingPlugin :: String -> Doc
whenLoadingPlugin name =
    fsep $ pwords "loading plugin" ++ [text name]

checkingSubtypingConstraint :: Value -> Value -> Doc
checkingSubtypingConstraint w v =
    fsep $ pwords "the inferred type" ++ [pretty w] ++
	    pwords "is not a subtype of" ++ [pretty v]

doingNothingMuch :: Doc
doingNothingMuch = fsep $ pwords "failed miserably"

-- Error messages ---------------------------------------------------------

variablesNotEqualError :: Name -> Name -> ErrorFun
variablesNotEqualError x y e =
    e	{ errWhy = fsep $ map text
			[ "identifier"
			, x
			, "is not equal to"
			, y
			]
	}

dontCareNotEqualError :: ErrorFun
dontCareNotEqualError e =
    e	{ errWhy = fsep $ pwords "_ is not equal to anything"
	}

sortsNotEqualError :: SortValue -> SortValue -> ErrorFun
sortsNotEqualError s1 s2 e =
    e	{ errWhy = fsep	[ pretty s1
			, text "is not equal to"
			, pretty s2
			]
	}

sortsNotLeqError :: SortValue -> SortValue -> ErrorFun
sortsNotLeqError s1 s2 e =
    e	{ errWhy = fsep	[ pretty s1
			, text "is not a subsort of"
			, pretty s2
			]
	}

notAtAllEqualError :: Value -> Value -> ErrorFun
notAtAllEqualError u v e =
    e	{ errWhy = fsep	[ text "the value"
			, pretty u
			, text "is not equal to"
			, pretty v
			]
	}

variableNotBoundError :: Name -> ErrorFun
variableNotBoundError x e =
    e	{ errWhy = fsep $
		    pwords "the variable"
		    ++ [text x] ++
		    pwords "has not been bound"
	}

notAFunctionError :: Term -> Value -> ErrorFun
notAFunctionError e v m =
    m   { errWhy = fsep $
		    pwords "it has type"
		    ++ [pretty v] ++
		    pwords "which is not a function type"
	, errWhat = fsep $
		    pretty e :
		    pwords "is not a function"
	}

notATypeError :: TypedTerm -> Value -> ErrorFun
notATypeError e v m =
    m	{ errWhy = fsep $
		    pwords "it has type"
		    ++ [pretty v] ++
		    pwords "which is not a sort"
	, errWhat = fsep $
			pretty e :
			pwords "is not a type"
	}

notASortError :: TypedTerm -> Value -> ErrorFun
notASortError e v m =
    m	{ errWhy = fsep $ pwords "evaluates to" ++ [pretty v] ++
			  pwords "which is not a sort"
	, errWhat = fsep $ [pretty e] ++ pwords "is not a sort"
	}

cannotInferError :: Term -> ErrorFun
cannotInferError e m =
    m	{ errWhat = fsep $ pwords "could not infer the type of"
			    ++ [pretty e]
	, errWhy = fsep $ pwords "it's not an inferrable thing (e.g. a lambda, constructor or pair)"
	}

lambdaHasFunctionTypeError :: Term -> Value -> ErrorFun
lambdaHasFunctionTypeError e v m =
    m	{ errWhat = fsep $ pretty e : pwords "does not have type" ++ [pretty v]
	, errWhy = fsep $ pwords "it has a function type"
	}

pairHasSigmaTypeError :: Term -> Value -> ErrorFun
pairHasSigmaTypeError e v m =
    m	{ errWhat = fsep $ pretty e : pwords "does not have type" ++ [pretty v]
	, errWhy = fsep $ pwords "it has a sigma type"
	}

notDefinedInternalError :: Name -> ErrorFun
notDefinedInternalError x m =
    m	{ errWhat   = fsep $ pwords "there was an internal error"
	, errWhy    = fsep $ pwords $ "the definition of " ++ x ++ " is missing"
	}

repeatedVariableError :: Name -> ErrorFun
repeatedVariableError x e =
    e	{ errWhat   = fsep $ pwords $ "the variable " ++ x ++ " is shadowed"
	}

pluginNotFoundError :: String -> ErrorFun
pluginNotFoundError name e =
    e	{ errWhat   = fsep $ pwords "failed to call plugin"
	, errWhy    = fsep $ pwords "the plugin" ++ [text name] ++
			     pwords "does not exist"
	}

commandLineParseError :: String -> ErrorFun
commandLineParseError err e =
    e	{ errWhy    = text err
	, errCode   = OptionError
	}

duplicateSigmaRuleError :: SortValue -> SortValue -> ErrorFun
duplicateSigmaRuleError v w e =
    e	{ errWhat   = fsep $ pwords "overlapping typing rules for Sigma"
	, errWhy    = fsep $ pwords "a rule for" ++ [sep [ pretty v <+> text ","
							 , nest 2 $ pretty w
							 ]
						    ] ++
			     pwords "has already been defined"
	}

duplicatePiRuleError :: SortValue -> SortValue -> ErrorFun
duplicatePiRuleError v w e =
    e	{ errWhat   = fsep $ pwords "overlapping typing rules for Pi"
	, errWhy    = fsep $ pwords "a rule for" ++ [sep [ pretty v <+> text ","
							 , nest 2 $ pretty w
							 ]
						    ] ++
			     pwords "has already been defined"
	}

piRuleMustHaveOneUserSortError :: SortValue -> SortValue -> ErrorFun
piRuleMustHaveOneUserSortError v w e =
    e	{ errWhat   = fsep $ pwords "invalid typing rule for Pi"
	, errWhy    = fsep $ pwords "one of" ++ [pretty v] ++
			     pwords "and" ++ [pretty w] ++
			     pwords "should be a user defined sort"
	}

emptySetError :: ErrorFun
emptySetError e =
    e	{ errWhat   = fsep $ pwords "sets must be non-empty"
	, errWhy    = fsep $ pwords "that's the way of the world"
	}

cannotReconstructTopError :: ErrorFun
cannotReconstructTopError e =
    e	{ errWhat   = fsep $ pwords "term reconstruction failed"
	, errWhy    = fsep $ pwords "cannot reconstruct a term from the value Top"
	}

