{-# OPTIONS -fglasgow-exts #-}
{-| The parser produces concrete syntax. This is translated to "InternalSyntax"
    before type checking.  

The declarations of the concrete syntax differ considerably from the
ones of the internal syntax.  Also, concrete syntax has explicit
parentheses and infix operators in a left associative way.  The
translation to internal syntax, reassociates infix expressions
according to their fixities.  

Fixities: Could be connected to the declaration of an identifier.
Their scope could be the scope of the identifier.  In the extreme,
this could be taken to
@
  \ (infixl 6 +) -> x + y
@
But I am not sure this is necessary.
-}

module Syntax.Concrete where

import Data.Generics (Typeable, Data)

import Syntax.Position

-- Concrete syntax --------------------------------------------------------

type Type = Term

data Name = Name Range String | DontCare

dontCare = DontCare

data Term   = App Term Term                     -- ^ . @t t'@
	    | HApp Term Term			-- ^ . @t {t'}@ hidden application
	    | Id Name				-- ^ . @x@
	    | Lam Range Pattern Term            -- ^ . @\ p -> t@
	    | Hash Range Universe               -- ^ . @#n | Set | Type@
	    | ISet Range			-- ^ . impredicative set
	    | Set Range				-- ^ . @Set@ (if universe polymorphism)
	    | Prop Range			-- ^ . @Prop@ (if universe polymorphism)
	    | Type Range			-- ^ . @Type@ (if universe polymorphism)
	    | Pi Range Pattern Term Term        -- ^ . @(p::t) -> t'@
            | CallPlugin Range String [Term]	-- ^ . @name-plugin(t1,...,tn)@
	    | Let Range [Decl] Term             -- ^ . @let ds in t@
	    | RewriteIn Range [Term] Term	-- ^ . @rewrite eqs in t@
	    | Paren Range Term                  -- ^ . @( t )@
	    | InfixApp Term Name Term		-- ^ . @t `x` t' | t + t'@
	    | Sigma Range Name Term Term	-- ^ . @{x:A; B}@
	    | Pair Range Term Term		-- ^ . @(t, u)@
	    | Project Range Term Projection	-- ^ . @p.i@
	    | Struct Range [(Name, [Pattern], [Pattern], Term)]	-- ^ . @struct f1 {xs} ys = e1; ..; fn = en@
	    | Meta Range

data Typing = Typing Term Term

data Universe = Univ Int

data Projection = Proj Range Int | Field Name

-- | Patterns are nested, made up from constructors and variables
data Pattern    = VarP Name				 -- ^ . @x@
                | ConP Range Name [Pattern] [Pattern]    -- ^ . @c ps@
                | WildP Range                            -- ^ . @_@
                | PairP Range Pattern Pattern            -- ^ . @(p1,p2)@
                | InfixConP Pattern Name Pattern -- ^ . @p1 : p2@

type Program	= [Decl]

-- | We currently have two forms of definitions. Normal definitions and
--   definitions of new constants.

data Decl	= TypeSig Range Name Telescope Type
		    -- ^ . @x :: t@
		| FunClause Range Name Clause
		    -- ^ . @f p0 .. pn = e@
		| Infix Range Assoc Precedence [Name]
                    -- ^ . @infix 1 == | infixl 2 `app` | infixr 3 =>@
		| Data Bool Range Name Telescope Telescope Type [Constructor]
                    -- ^ . @[sized] data D (xs::As) :: s where cs::Cs@
                    -- Bool=True iff sized
		| Signature Range Name Telescope Telescope Type [Name] [Constructor]
		| Rewrite Range [Term]
		| Mutual Range [Decl]
		| Abstract Range [Decl]

data Clause = Clause Range [Pattern] [Pattern] WithClause

data WithClause = NoWith Term
		| With [Term] [Clause]

data Rule   = PiRule Range Term Term Term
                -- ^ . @Pi Set Prop -> Prop@
	    | SigmaRule Range Term Term Term
                -- ^ . @Sigma Set Prop -> Prop@

data Constructor = Constructor Range Name Telescope Type
                    -- ^ . @c :: C@

constrName :: Constructor -> Name
constrName (Constructor _ x _ _) = x

type Telescope	= [(Name, Type)]

data Assoc	= NonAssoc | LeftAssoc | RightAssoc
    deriving (Typeable, Data)

{- | Precedences are positive integers.  
Juxtaposition has precedence +infinity. The default precedence is 9. -}
type Precedence = Int
type Fixity	= (Assoc, Precedence)

defaultFixity :: Fixity
defaultFixity = (LeftAssoc, 9)

--type LocalDefinition = Definition

--type Telescope = [(Name, Term)]

-- Positions --------------------------------------------------------------

instance HasRange Term where
    getRange (App t u)		    = fuseRange t u
    getRange (Id x)		    = getRange x
    getRange (Lam r _ _)	    = r
    getRange (Hash r _)		    = r
    getRange (Pi r _ _ _)	    = r
    getRange (Pair r _ _)	    = r
    getRange (Project r _ _)	    = r
    getRange (Sigma r _ _ _)	    = r
    getRange (CallPlugin r _ _)	    = r
    getRange (Let r _ _)	    = r
    getRange (RewriteIn r _ _)	    = r
    getRange (Paren r _)	    = r
    getRange (Set r)		    = r
    getRange (Prop r)		    = r
    getRange (Type r)		    = r
    getRange (ISet r)		    = r
    getRange (Meta r)		    = r
    getRange (Struct r _)	    = r
    getRange (InfixApp t _ u)	    = fuseRange t u
    getRange (HApp t u)		    = fuseRange t u

instance HasRange Decl where
    getRange (TypeSig r _ _ _)		= r
    getRange (FunClause r _ _)		= r
    getRange (Infix r _ _ _)		= r
    getRange (Rewrite r _)		= r
    getRange (Mutual r _)		= r
    getRange (Abstract r _)		= r
    getRange (Data _ r _ _ _ _ _)	= r
    getRange (Signature r _ _ _ _ _ _)	= r

instance HasRange WithClause where
    getRange (NoWith e)		    = getRange e
    getRange (With ps cs)	    = fuseRange ps cs

instance HasRange Clause where
    getRange (Clause r _ _ _)	    = r

instance HasRange Pattern where
    getRange (VarP x)		    = getRange x
    getRange (ConP r _ _ _)	    = r
    getRange (PairP r _ _)	    = r
    getRange (InfixConP p1 _ p2)    = fuseRange p1 p2
    getRange (WildP r)		    = r

instance HasRange Constructor where
    getRange (Constructor r _ _ _) = r

instance HasRange Projection where
    getRange (Proj r _)	= r
    getRange (Field x)  = getRange x

instance HasRange Name where
    getRange (Name r _) = r
    getRange DontCare	= noRange

instance Show Name where
    show (Name _ x) = x
    show DontCare   = "_"

instance Eq Name where
    Name _ x == Name _ y = x == y
    DontCare == DontCare = True
    _ == _		 = False

