{-# OPTIONS -fglasgow-exts -fallow-overlapping-instances #-}

module Syntax.Pretty
    ( module Syntax.Pretty
    , module Utils.Pretty
    , module Text.PrettyPrint
    ) where

import Prelude hiding	    ( print, putStr, putStrLn )
import IO.Output

import Control.Monad
import Control.Monad.State
import Data.List
import Data.IORef
import System.IO.Unsafe ( unsafePerformIO )
import Text.PrettyPrint

import Syntax.Internal
import Syntax.Values
import Syntax.Position
import TypeChecker.Context
import TypeChecker.Abstractions ( vApps, vApply )

import Utils.Names	    ( noNamesUsed, fresh, UsedNames )
import Utils.Monads
import Utils.Pairs
import Utils.Pretty

-- | We want the fresh names to be printed with numbers
freshPrettyName "_" = return DontCare
freshPrettyName x =
    do	Name r x' n <- fresh x
	return $ Name r (x' ++ show n) n

-- Show instances ---------------------------------------------------------

instance Show Term		    where   show = show . pretty
instance Show Pattern		    where   show = show . pretty
instance Show Clause		    where   show = show . pretty
instance Show Projection	    where   show = show . pretty
instance Show Value		    where   show = show . pretty
instance Show Type		    where   show = show . pretty
instance Show SortValue		    where   show = show . pretty
instance Show ClauseValue	    where   show = show . pretty
instance Show DefValue		    where   show = show . pretty
instance Show Universe		    where   show = show . pretty
instance Show Program		    where   show = show . pretty
instance Show Definition	    where   show = show . pretty
instance Show ProjectionValue	    where   show = show . pretty
instance Show ContextEntry	    where   show = show . pretty
instance Show MetaRef		    where   show = show . pretty
instance Show Name		    where   show = show . pretty

instance Pretty Doc where
    pretty = id

instance Pretty a => Pretty [a] where
    pretty = brackets . fsep . punctuate comma . map pretty

-- Pretty ranges
instance Pretty Range where
    pretty = text . show

-- Pretty names
instance Pretty Name where
    pretty (Name _ x n) = text $ x -- ++ "_" ++ show n
    pretty DontCare	= text "_"
--    pretty = text . nameString

-- Pretty MetaRefs
instance Pretty MetaRef where
    pretty m =
	case unsafePerformIO $ readIORef m of
	    Open	    -> text "?"
	    Instantiated v  -> text "?" <> braces (pretty v)

-- Pretty programs --------------------------------------------------------

instance Pretty Program where
    pretty prog = vcat $ map pretty prog

-- Pretty definitions -----------------------------------------------------

getHidden :: Int -> Term -> (Telescope, Term)
getHidden 0 t			= ([], t)
getHidden n (Pi _ (VarP x) a b)	= ((x,a):) -*- id $ getHidden (n - 1) b
getHidden _ _			= error "impossible: Syntax.Pretty.getHidden"

instance Pretty Definition where
    pretty (NormalDef _ name h _ t' cs) =   -- TODO: pretty print hidden
	vcat
	$ sep [ prettyName name <+> prettyTelescope tel <+> text ":"
	      , nest 2 $ pretty t
	      ]
	: map pretty cs
	where
	    (tel,t) = getHidden h t'
    pretty (Synonym _ x t) = sep [ prettyName x <+> text "="
				 , nest 2 $ pretty t
				 ]
    pretty (NewAxiom _ name h t) =
	sep [ prettyName name <+> prettyTelescope tel <+> text ":"
	    , nest 2 $ pretty t'
	    ]
	where
	    (tel,t') = getHidden h t
    pretty (NewData sized _ name htel tel t cs) =
	sep [ fsep  [ ((\ s -> if sized then text "sized" <+> s else s)
                      (text "data")) <+> prettyName name
		    , nest 4 $ prettyHTelescope htel
		    , nest 4 $ prettyTelescope tel
		    , nest 4 $ text ":"
		    , nest 4 $ pretty t
		    , nest 4 $ text "where"
		    ]
	    , nest 2 $ vcat $ map pretty cs
	    ]
    pretty (NewSig _ name htel tel t ps cs) =
	sep [ fsep  [ text "signature" <+> prettyName name
		    , nest 4 $ prettyHTelescope htel
		    , nest 4 $ prettyTelescope tel
		    , nest 4 $ text ":" <+> pretty t
		    , nest 2 $ text "extends" <+> fsep (punctuate comma $ map pretty ps)
		    , nest 2 $ text "where"
		    ]
	    , nest 2 $ vcat $ map pretty cs
	    ]
    pretty (Rewrite _ es) = text "rewrite" <+> vcat (map pretty es)
    pretty (AbstractDefs _ ds) = sep [ text "abstract"
			       , nest 2 $ vcat $ map pretty ds
			       ]
    pretty (Mutual _ ds) = sep [ text "mutual"
			       , nest 2 $ vcat $ map pretty ds
			       ]

instance Pretty LocalDefinition where
    pretty (LocalDef _ name mt _ e) =
	vcat	[ maybe empty (\t ->
		    sep [ prettyName name <+> text ":"
			, nest 2 $ pretty t
			]
		    ) mt
		, sep	[ prettyName name
			, nest 2 $ text "=" <+> pretty e
			]
		]

instance Pretty Constructor where
    pretty (Constructor _ name tel t) =
	sep [ prettyName name <+> prettyTelescope tel <+> text ":"
	    , nest 2 $ pretty t
	    ]

-- Pretty telescopes ------------------------------------------------------

-- | Do abbr.
prettyTelescope tel = fsep $ map pr tel
    where
	pr (x,t) = parens $ sep [ prettyName x <+> text ":"
				, nest 2 $ pretty t
				]

prettyHTelescope tel = fsep $ map pr tel
    where
	pr (x,t) = braces $ sep [ prettyName x <+> text ":"
				, nest 2 $ pretty t
				]

-- Pretty clauses ---------------------------------------------------------

instance Pretty Clause where
    pretty (Clause _ x ps w) =
	sep [ fsep $ prettyName x : (map (nest 2) $ map (prettyPrec 10) ps)
	    , nest 2 $ pretty w
	    ]

instance Pretty WithClause where
    pretty (NoWith e)	= text "=" <+> pretty e
    pretty (With es cs) =
	vcat $ (text "with" <+> fsep (map pretty es)) :
		map (nest 2 . pretty) cs

instance Pretty ClauseValue where
    pretty (ClauseVal ps e) =
	sep [ fsep $ map (prettyPrec 1) ps ++ [nest 2 (text "->")]
	    , pretty e
	    ]

-- Pretty patterns --------------------------------------------------------

instance Pretty Pattern where
    prettyPrec _ (WildP _)	    = text "_"
    prettyPrec _ (VarP x)	    = prettyName x
    prettyPrec _ p@(PairP _ _ _)    =
	parens $ fsep $ punctuate comma $ pPair p
	    where
		pPair (PairP _ p q) = pretty p : pPair q
		pPair p		    = [pretty p]
    prettyPrec pr (ConP _ x hs ps)  =
	mparens (pr > 0 && not (null ps)) $
	    fsep    ( prettyName x
		    :  map (nest 2 . braces . pretty) hs
		    ++ map (nest 2 . prettyPrec 10) ps
		    )

-- Pretty projections -----------------------------------------------------

instance Pretty Projection where
    pretty (Proj _ n)	= text $ show n
    pretty (Field _ x)	= text x

-- Pretty terms -----------------------------------------------------------

data PrettyTerm
	= PApp Term Term
	| PHApp Term Term
	| PInfix Term Name Term
	| PName Name
	| PPrimitive String
	| PLam [Pattern] Term
	| PHash Universe
	| PPi [Pattern] Term Term
	| PFun Term Term
	| PPlugin String [Term]
	| PLet [Definition] Term
	| PRewrite [Term] Term
	| PTup [Term]
	| PProj Term Projection
	| PSigma [([Name], Term)]
	| PStruct [(Name, [Pattern], [Pattern], Term)]
	| PMeta Name
	| PNat Int
	| PList [Term]

prettyView :: Term -> PrettyTerm
prettyView e =
    case e of
	Con _ (Name _ "zero" _)	-> PNat 0
	App (Con _ (Name _ "suc" _)) e1
	    | PNat n <- prettyView e1	-> PNat (n + 1)
	Con _ (Name _ "nil" _) -> PList []
	App (App (Con _ (Name _ "::" _)) e1) e2
	    | PList es <- prettyView e2    -> PList (e1 : es)
	App (App (Var _ x) e1) e2
	    | isOperator x  -> PInfix e1 x e2
	App (App (Def _ x) e1) e2
	    | isOperator x  -> PInfix e1 x e2
	App (App (Con _ x) e1) e2
	    | isOperator x  -> PInfix e1 x e2
	App (App (Data _ x) e1) e2
	    | isOperator x  -> PInfix e1 x e2
	App e1 e2   -> PApp e1 e2
	HApp e1 e2  -> PHApp e1 e2
	Var _ x	    -> PName x
	Def _ x	    -> PName x
	Con _ x	    -> PName x
	Data _ x    -> PName x
	Sig _ x	    -> PName x
	Meta _ x    -> PMeta x
	Lam _ p e   -> case prettyView e of
			    PLam xs e'	-> PLam (p:xs) e'
			    _		-> PLam [p] e
	Hash _ u    -> PHash u
	ISet _	    -> PPrimitive "ISet"
	Set _	    -> PPrimitive "Set"
	Type _	    -> PPrimitive "Type"
	Prop _	    -> PPrimitive "Prop"
	Pi _ (WildP _) t e	-> PFun t e
	Pi _ p t e		->
		case prettyView e of
		    PPi ps t' e' | t == t'  -> PPi (p:ps) t e'
		    _			    -> PPi [p] t e
	CallPlugin _ _ x es	-> PPlugin x es
	Let _ ds e		-> PLet ds e
	RewriteIn _ e e'	-> PRewrite e e'
	Pair _ t u	-> case prettyView u of
			    PTup ts	-> PTup (t:ts)
			    _	-> PTup [t,u]
	Struct _ xs	-> PStruct xs
	Project _ t p   -> PProj t p
	Sigma _ DontCare a b ->
	    case prettyView b of
		PSigma tel  -> PSigma $ ([],a) : tel
		_	    -> PSigma [([],a), ([],b)]
	Sigma _ x a b ->
	    case prettyView b of
		PSigma tel@(([],_):_)	-> PSigma $ ([x],a) : tel
		PSigma ((xs, b') : tel)
		    | a == b'   -> PSigma ((x : xs, a):tel)
		    | otherwise -> PSigma (([x],a):(xs,b'):tel)
		_		-> PSigma [([x],a), ([],b)]
    where
	names "_" = []
	names x = [x]

inf = maxBound :: Int
appPrec	    = inf - 2
projPrec    = inf - 1
infixPrec   = 5

instance Pretty Term where

    prettyPrec p e =
	case prettyView e of
	    PApp e1 e2	-> 
		mparens (p > appPrec) $
		    sep [ prettyPrec appPrec e1
			, nest 2 $ prettyPrec (appPrec + 1) e2
			]

	    PHApp e1 e2	-> 
		mparens (p > appPrec) $
		    sep [ prettyPrec appPrec e1
			, nest 2 $ braces $ pretty e2
			]

	    PInfix e1 op e2 ->
		mparens (p > infixPrec) $
		    fsep [ prettyPrec (infixPrec + 1) e1
			 , prettyOp op
			 , prettyPrec (infixPrec + 1) e2
			 ]
	    PPrimitive x    -> text x
	    PName x	    -> prettyName x
	    PMeta x	    -> text "?"

	    PNat n	    -> text $ show n
	    PList es	    -> brackets $ fsep $ punctuate comma $ map pretty es

	    PLam ps e	->
		mparens (p>0) $
		    sep	[ sep [ text "\\" <> fsep (map pretty ps), text "->" ]
			, nest 2 $ pretty e
			]

	    PHash UZero		-> text "Set"
	    PHash (USucc UZero) -> text "Type"
	    PHash u		-> text "#" <> pretty u

	    PRewrite es e	-> sep [ text "rewrite" <+> vcat (map pretty es)
				       , text "in" <+> pretty e
				       ]

	    PProj t pr		->
		mparens (p > projPrec) $
		    prettyPrec projPrec t <> text "." <> pretty pr

	    (PFun _ _)	->
		mparens (p>0) $
		    fsep $ punctuate (text " ->")
			    (map (prettyPrec 1) $ funs e)
		where
		    funs a = case prettyView a of
				PFun a b    -> a : funs b
				_	    -> [a]

	    PPi ps a b		->
		mparens (p>0) $
		    sep [ parens (fsep $
				    (punctuate (text ",") $ map pretty ps)
				    ++	[ text ":"
					, nest 2 $ pretty a
					]
				    )
			    <+> text "->"
			, pretty b
			]

	    PSigma tel		->
		braces	$ fsep $ punctuate (text ";")
			$ map prettyTel tel
		where
		    prettyTel ([],a) = pretty a
		    prettyTel (xs,a) =
			fsep $ (punctuate (text ",") $ map pretty xs)
			    ++	[ text ":"
				, nest 2 $ pretty a
				]

	    PStruct xs	->
		mparens (p>0) $
		text "struct" <+> vcat [ sep [ prettyName x
						<+> fsep (map hid hps ++ map pretty ps)
						<+> text "="
					     , nest 2 $ pretty t
					     ]
				       | (x,hps,ps,t) <- xs
				       ]
		    where
			hid x = braces $ pretty x

	    PTup ts ->
		parens $ fsep $ punctuate (text ",") $ map pretty ts

	    PPlugin x es ->
		sep [ text $ x ++ "-plugin"
		    , nest 2 $ parens	$ fsep
					$ punctuate (text ",")
					$ map pretty es
		    ]

	    PLet ds e		->
		mparens (p>0) $
		    fsep    [ text "let" <+> vcat (map pretty ds)
			    , text "in" <+> pretty e
			    ]

prettyName x = mparens (isOperator x) $ pretty x
prettyOp x
    | isOperator x  = pretty x
    | otherwise	    = text "`" <> pretty x <> text "`"

-- Pretty universes -------------------------------------------------------

instance Pretty Universe where
    pretty u = pretty (uValue u)

-- Pretty values ----------------------------------------------------------

instance Pretty Type where
    prettyPrec p t = prettyPrec p (unEl t)

instance Pretty Value where
    prettyPrec p v = flip evalState noNamesUsed $ prettyM p $ unsafeRefresh v
	where
	    unsafeRefresh v =
		case vAppView v of
		    VAVMeta m vs    ->
			case unsafePerformIO $ readIORef m of
			    Open	    -> v
			    Instantiated v' -> ur $ vApps v' vs
		    _	-> case vView v of
			    VApp v1 v2	    -> vApp (ur v1) (ur v2)
			    VPi w f	    -> vPi (urT w) (urAT f)
			    VLam f	    -> vLam (urAV f)
			    VPair v1 v2	    -> vPair (ur v1) (ur v2)
			    VProj v p	    -> vProj (ur v) p
			    VSigma w f	    -> vSigma (urT w) (urAT f)
			    _		    -> v
		where
		    ur = unsafeRefresh
		    urT = typeMap unsafeRefresh
		    urAV = absMap ur
		    urAT = absMap urT

	    inst f v = vLam f `vApply` v
	    instT f = inst (fmap unEl f)

	    prettyM :: Int -> Value -> State UsedNames Doc
	    prettyM p v
		| Just n <- maybeNat v	    = return $ text (show n)
		| Just vs <- maybeList v    =
		    do	ds <- mapM (prettyM 0) vs
			return $ brackets $ fsep $ punctuate comma ds
	    prettyM p v =
		case vAppView $ vDefToCon v of
		    VAVId x vs@[_,_]
			| isOperator x	->
			    do	[d1,d2] <- mapM (prettyM 5) vs
				return $ mparens (p > 4)
				       $ fsep [ d1, prettyOp x, d2 ]
		    VAVId x vs    -> pApp (prettyName x) vs
		    VAVBId n vs   -> pApp (text $ "@" ++ show n) vs
		    VAVDef _ _ _    -> error "impossible: Syntax.Pretty.prettyM"
		    VAVCon x h _ vs
			| isOperator x && length vs - h == 2 ->
			    do	[d1,d2] <- mapM (prettyM 5) $ drop h vs
				return $ mparens (p > 4)
				       $ fsep [ d1, prettyOp x, d2 ]
			| otherwise -> 
			    do	let name = prettyName x
				pHApp name hvs vs'
			where (hvs, vs') = splitAt h vs
		    VAVMeta m vs    -> pApp (text "?") vs
		    VAVSort s	    -> return $ pretty s

		    VAVPair _ _ ->
			do  ds <- pPair v
			    return $ parens $ fsep $ punctuate comma ds
			where
			    eq x d = sep [ text x <+> text "=", nest 2 d ]
			    pPair v = case vView v of
				VPair v w -> do	dv <- prettyM 0 v
						ds <- pPair w
						return (dv:ds)
				_	    -> (:[]) <$> prettyM 0 v

		    VAVProj v p vs ->
			do  dw <- prettyM projPrec w
			    pApp (dw <> text "."
				     <> text f) ws
			where
			    (w,ws,f) = projectionNames v p vs
			    projectionNames v VFst vs = (v, vs, "1")
			    projectionNames v VSnd vs = (v, vs, "2")
			    projectionNames v (VField h x) vs = (v, drop h vs, x)

		    VAVSigma _ _    ->
			do  ds <- pSigma v
			    return $ braces $ fsep $ punctuate semi ds
			where
			    named x d
				| nameString x == "_" = d
				| otherwise	      =
				    sep [ pretty x <+> text ":"
					, nest 2 d
					]

			    pSigma v =
				case vView v of
				    VSigma w f ->
					do  x' <- freshPrettyName $ absName f
					    dw <- named x' <$> prettyM 0 (unEl w)
					    ds <- pSigma $ f `instT` vId x'
					    return (dw:ds)
				    _	-> (:[]) <$> prettyM 0 v

		    VAVStruct xs ->
			do  ds <- mapM pStruct xs
			    return $ mparens (p>0) $
				sep [ text "struct"
				    , nest 2 $ vcat ds
				    ]
			where
			    pStruct (x,(h,v)) =
				do  (args,v') <- hiding h v
				    d   <- prettyM 0 v'
				    return $ sep [ prettyName (dummyName x) <+> fsep args <+> text "="
						 , nest 2 d
						 ]
			    hiding 0 v = return ([],v)
			    hiding n v =
				do  x <- aName (vView v)
				    (args,v') <- hiding (n - 1) $ v `vApply` vId x
				    return (braces (prettyName x) : args, v')

			    aName (VLam f) = freshPrettyName $ absName f
			    aName _	   = freshPrettyName "h"

		    VAVPi _ _ ->
			do  ds <- pPi v
			    return $ mparens (p>0) $
				fsep $ punctuate (text " ->") ds
			where
			    pPi v =
				case vView v of
				    VPi v f
					| absName f == "_" ->
					    do	dv <- prettyM 1 (unEl v)
						ds <- pPi $ f `instT` vId dontCare
						return (dv:ds)
					| otherwise ->
					    do  dv  <- prettyM 0 (unEl v)
						x   <- freshPrettyName $ absName f
						ds  <- pPi $ f `instT` vId x
						let dv' = parens $ sep
							[ prettyName x <+> text ":"
							, nest 2 dv
							]
						return (dv':ds)
				    _	-> (:[]) <$> prettyM 0 v

		    VAVLam _	->
			do  (xs, df) <- pLam v
			    return $ mparens (p>0) $
				sep [ text "\\" <> fsep (map prettyName xs ++ [text "->"])
				    , nest 2 df
				    ]
			where
			    pLam v =
				case vView v of
				    VLam f  ->
					do  x <- freshPrettyName $ absName f
					    (xs,df) <- pLam $ f `inst` vId x
					    return (x:xs, df)
				    _	->
					do  dv <- prettyM 0 v
					    return ([], dv)
		    
		    VAVTop  -> return $ text "Top" -- <+> text ":" <+> pretty t

		where
		    pApp' d ds = mparens par $ sep [ d, nest 2 $ fsep ds ]
			where
			    par = not (null ds) && p > 9

		    pHApp d hs vs =
			do  hds <- mapM (prettyM 0) hs
			    vds <- mapM (prettyM 10) vs
-- 			    return $ pApp' d (map braces hds ++ vds)
			    return $ pApp' d vds

		    pApp d vs =
			do  ds <- mapM (prettyM 10) vs
			    return $ pApp' d ds

	    maybeNat v =
		case vAppView $ vDefToCon v of
		    VAVCon (Name _ "zero" _) _ _ [] -> return 0
		    VAVCon (Name _ "suc" _) _ _ [n] -> liftM (1+) $ maybeNat n
		    _			    -> Nothing

	    maybeList v =
		case vAppView $ vDefToCon v of
		    VAVCon (Name _ "nil" _) _ _ [_] -> return []
		    VAVCon (Name _ "::" _) _ _ vs
			| xs:x:_ <- reverse vs	-> liftM (x:) $ maybeList xs
		    _				-> Nothing

	    annotate p t d = d -- mparens True $ sep [ d, text ":", pretty t ]

instance Pretty ContextEntry where
    prettyPrec p (Abstract t)  = mparens (p>0) $ text "(abstract)" <+> pretty t
    prettyPrec p (Defined v t) = mparens (p>0) $ text "(defined)" <+> pretty v

instance Pretty SortValue where
    pretty VSISet		    = text "ISet"
    pretty (VSUniv UZero)	    = text "Set"
    pretty (VSUniv (USucc UZero))   = text "Type"
    pretty (VSUniv u)		    = text "#" <> pretty u
    pretty VSProp		    = text "Prop" -- <> braces (prettyPrec 0 i)

instance Pretty ProjectionValue where
    pretty VFst		= text "1"
    pretty VSnd		= text "2"
    pretty (VField _ x)	= text x

instance Pretty DefValue where
    pretty (DefVal h ar fv cl t) =
	text "DefValue" <+>
	    fsep[ brackets $ pretty h
		, pretty ar
		, pretty fv
		, pretty cl
		, prettyPrec (projPrec + 1) t
		]

instance Pretty ActualDefinition where
    pretty def =
	case def of
	    AbstrDef cl	-> text "abstract" <> pr cl
	    ConcrDef cl -> pr cl
	where
	    pr = brackets . fsep . punctuate comma . map pretty

-- Pretty ints ------------------------------------------------------------

instance Pretty Int where
    pretty = text . show

-- Helpers ----------------------------------------------------------------

mparens True	= parens
mparens _	= id

